// ---------------------------------------------------------------------
// This source file is (C) 2010-2013 Sanders Technology http://www.sanderstechnology.com
// Please review the copyright terms in the Conditions.txt file accompanying this source code
// Author:  Rob Sanders
// Email:   rob.sanders@gmail.com
// Created: Thursday, 10 February 2011, Last Modified: 23 October, 2012
// ---------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Globalization;
using System.Configuration;
using Common.Logging;

namespace Common
{
    /// <summary>
    /// Handles Retrieving Configuration Values from the App.Config (or as appropriate)
    /// </summary>
    public static class ConfigManager
    {
        #region Get Value Functions

        /// <summary>
        /// Retrieves a Connection String from the App.Config/Web.Config
        /// </summary>
        /// <param name="connectionName"></param>
        /// <returns></returns>
        public static string GetConnectionString(string connectionName)
        {
            var configSetting = ConfigurationManager.ConnectionStrings[connectionName];
            
            if(configSetting == null || String.IsNullOrEmpty(configSetting.ConnectionString))
            {
                WriteLogMessage(connectionName);
                return String.Empty;
            }

            return configSetting.ConnectionString;
        }

        /// <summary>
        /// Try to get a string value from the App.config
        /// </summary>
        /// <param name="keyName">the key name to lookup</param>
        /// <returns>String.Empty (on error) or the value specified in the config</returns>
        public static string GetStringValue(string keyName)
        {
            string configSetting = System.Configuration.ConfigurationManager.AppSettings[keyName];

            if (configSetting == null)
            {
                WriteLogMessage(keyName);
                return String.Empty;
            }

            return configSetting;
        }

        /// <summary>
        /// Try to get a string value from the App.config
        /// </summary>
        /// <param name="keyName">the key name to lookup</param>
        /// <param name="defaultValue">A value to return if the key or value is not found</param>
        /// <param name="logError">Optionally log the missing key</param>
        /// <returns>String.Empty (on error) or the value specified in the config</returns>
        public static string GetStringValue(string keyName, string defaultValue, bool logError)
        {
            string value = System.Configuration.ConfigurationManager.AppSettings[keyName];

            if (value == null)
            {   
                if (logError)
                    WriteLogMessage(keyName, String.Format(CultureInfo.CurrentCulture, "is missing; using default value, {0}", defaultValue));

                return defaultValue;
            }

            return value;
        }

        /// <summary>
        /// Try to get a string value from the App.config
        /// </summary>
        /// <param name="keyName">the key name to lookup</param>
        /// <param name="defaultValue">A value to return if the key or value is not found</param>
        /// <returns>String.Empty (on error) or the value specified in the config</returns>
        public static string GetStringValue(string keyName, string defaultValue)
        {
            string value = GetStringValue(keyName);

            if (value.Length == 0)
            {
                WriteLogMessage(keyName, String.Format(CultureInfo.CurrentCulture, "using default value, {0}", defaultValue));
                return defaultValue;
            }

            return value;
        }

        /// <summary>
        /// Try to get an int value from the App.config
        /// </summary>
        /// <param name="keyName"></param>
        /// <returns>int.MinValue (on error) or the value specified in the config</returns>
        public static int GetIntValue(string keyName)
        {
            int retValue = int.MinValue;
            if (!int.TryParse(System.Configuration.ConfigurationManager.AppSettings[keyName], out retValue))
            {
                WriteLogMessage(keyName);
            }
            return retValue;
        }

        /// <summary>
        /// Try to get an int value from the App.config
        /// </summary>
        /// <param name="keyName"></param>
        /// <param name="defaultValue">the default to use</param>
        /// <returns>int.MinValue (on error) or the value specified in the config</returns>
        public static int GetIntValue(string keyName, int defaultValue)
        {
            int retValue = int.MinValue;
            if (!int.TryParse(System.Configuration.ConfigurationManager.AppSettings[keyName], out retValue))
            {
                WriteLogMessage(keyName, String.Format(CultureInfo.CurrentCulture, "using default value, {0}", defaultValue));
                return defaultValue;
            }
            return retValue;
        }

        /// <summary>
        /// Try to get an int value from the App.config
        /// </summary>
        /// <param name="keyName"></param>
        /// <returns>int.MinValue (on error) or the value specified in the config</returns>
        public static bool GetBooleanValue(string keyName)
        {
            bool retValue = false;
            if (!bool.TryParse(System.Configuration.ConfigurationManager.AppSettings[keyName], out retValue))
            {
                WriteLogMessage(keyName);
            }
            return retValue;
        }

        /// <summary>
        /// Try to get an int value from the App.config
        /// </summary>
        /// <param name="keyName"></param>
        /// <returns>int.MinValue (on error) or the value specified in the config</returns>
        public static bool GetBooleanValue(string keyName, bool defaultValue)
        {
            bool retValue;
            if (!bool.TryParse(System.Configuration.ConfigurationManager.AppSettings[keyName], out retValue))
            {
                WriteLogMessage(keyName);
                return defaultValue;
            }
            return retValue;
        }

        #endregion

        #region Private Functions

        /// <summary>
        /// Write to the Event Log
        /// </summary>
        /// <param name="keyName">Key Name</param>
        private static void WriteLogMessage(string keyName)
        {
            string logMessage = String.Format(CultureInfo.CurrentCulture, "Missing or in valid key {0} in the service app.config file", keyName);
            EventLogProvider.GetInstance().WriteConfigLog("Application", logMessage);
        }

        /// <summary>
        /// Write to the Event Log
        /// </summary>
        /// <param name="keyName">Key Name</param>
        /// <param name="message">Additional Message</param>
        private static void WriteLogMessage(string keyName, string message)
        {
            string logMessage = String.Format(CultureInfo.CurrentCulture, "Missing or in valid key {0} in the service app.config file, {1}", keyName, message);
            EventLogProvider.GetInstance().WriteConfigLog("Application", logMessage);
        }

        #endregion
    }
}
