﻿// ---------------------------------------------------------------------
// This source file is (C) 2010-2013 Sanders Technology http://www.sanderstechnology.com
// Please review the copyright terms in the Conditions.txt file accompanying this source code
// Author:  Rob Sanders
// Email:   rob.sanders@gmail.com
// Created: Thursday, 10 February 2011, Last Modified: 23 October, 2012
// ---------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Reflection;
using System.Threading;

namespace Common.Logging
{
    public class FlatFileProvider : BaseLogProvider, ILogProvider, IDisposable
    {
        #region Private Variables

        private static FlatFileProvider _instance;
        private static StreamWriter _writer;
        private static string _fileName;
        private static string _path;
        private bool _ended = false;
        private bool disposed = false;
        private static object _sync = new object();
        //private static object _creationSync = new object();
        private static bool _isInitialized = false;
        private static string _sourceName = ConfigManager.GetStringValue("LogFileSourceName", "DMS");

        #endregion

        #region Constructors

        public FlatFileProvider()
        {            
            InitializeLog();
        }     

        public FlatFileProvider(string fileName)
        {
            _fileName = fileName;

            InitializeLog();
        }

        public FlatFileProvider(string path, string fileName)
        {
            _fileName = fileName;
            _path = path;

            InitializeLog();
        }

        #endregion

        #region Public Properties

        /// <summary>
        /// Returns the Log File Name
        /// </summary>
        public string LogFileName
        {
            get
            {
                return _fileName;
            }
        }

        /// <summary>
        /// Returns the location of the Log File
        /// </summary>
        public string LogFileLocation
        {
            get
            {
                return _path;
            }
        }


        #endregion

        #region Instance Handling

        public static FlatFileProvider GetInstance()
        {
            Monitor.Enter(_sync);
            try
            {
                if (_instance == null)
                {
                    _instance = new FlatFileProvider();
                }
                return _instance;
            }
            finally
            {
                Monitor.Exit(_sync);
            }            
        }

        public static FlatFileProvider GetInstance(string fileName)
        {
            Monitor.Enter(_sync);
            try
            {
                if (_instance == null)
                {
                    _instance = new FlatFileProvider(fileName);
                }
                return _instance;
            }
            finally
            {
                Monitor.Exit(_sync);
            }            
        }

        #endregion

        #region Public Functions

        public void ChangeLogFile(string name)
        {
            Assembly thisAssembly = Assembly.GetExecutingAssembly();
            string path = thisAssembly.Location.Replace(thisAssembly.ManifestModule.Name, String.Empty);
            ChangeLogFile(path, name);
        }

        public void ChangeLogFile(string path, string name)
        {
            Cleanup();
            _writer = new StreamWriter(Path.Combine(path, name));
            _writer.AutoFlush = true;
        }

        public bool Write(string message)
        {
            if (_writer == null)
            {
                InitializeLog();
            }
            _writer.Write(message);
            return true;
        }

        public bool WriteLine(string message)
        {
            return WriteLine(message, true);
        }

        public bool WriteLine(string message, bool includeFormatting)
        {
            if (_writer == null)
            {
                InitializeLog();
            }
            
            if (includeFormatting)
            {
                _writer.WriteLine(FormatMessage(message));
            }
            else
            {
                _writer.WriteLine(message);
            }
            return true;
        }

        public bool Write(LogDetails details)
        {
            string message = details.Message;
            return Write(message);
        }

        public bool WriteLine(LogDetails details)
        {
            string message = details.Message;
            return WriteLine(message);
        }


        public void Cleanup()
        {
            if (_writer != null)
            {
                _writer.Close();
            }
        }

        #endregion

        #region Private Functions

        private void InitializeLog()
        {                       
            if (_isInitialized)
                return;

            if (String.IsNullOrEmpty(_fileName))
            {
                _fileName = String.Format("{0}-{1}.log", DateTime.Now.ToString("yyyyMMdd-hhmmss"), _sourceName);
            }
            if (String.IsNullOrEmpty(_path))
            {
                //Assembly thisAssembly = Assembly.GetExecutingAssembly();
                //_path = thisAssembly.Location.Replace(thisAssembly.ManifestModule.Name, String.Empty);
                _path = Path.GetTempPath();
            }

            _writer = new StreamWriter(Path.Combine(_path, _fileName), false);
            _writer.AutoFlush = true;
            _isInitialized = true;           
        }

        #endregion

        #region Idisposable Members

        /// <summary>
        /// Handle Object Disposal
        /// </summary>
        public void Dispose()
        {
            if (!this.disposed)
            {
                this.Dispose(true);
                GC.SuppressFinalize(this);
                this.disposed = true;
            }
        }

        /// <summary>
        /// Ensure proper cleanup
        /// </summary>
        protected void Dispose(bool disposing)
        {
            if (!_ended)
            {
                try
                {
                    Cleanup();
                }
                catch (Exception ex)
                {
                    EventHandlers.HandleExceptionEvent(ex);
                    _ended = true;
                }
            }
        }

        #endregion
    }
}
