﻿// ---------------------------------------------------------------------
// This source file is (C) 2010-2013 Sanders Technology http://www.sanderstechnology.com
// Please review the copyright terms in the Conditions.txt file accompanying this source code
// Author:  Rob Sanders
// Email:   rob.sanders@gmail.com
// Created: Thursday, 10 February 2011, Last Modified: 23 October, 2012
// ---------------------------------------------------------------------
#region Using Directives
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Collections.Concurrent;
using Common;
#endregion

namespace ThreadManager
{
    /// <summary>
    /// A Generic Implementation of a Fully Featured Thread Pool Wrapper
    /// </summary>
    public class ThreadPoolManager : IDisposable
    {
        #region Private (Instance) Variables

        private int _totalThreads    = 0;
        private int _executedThreads = 0;
        private int _abortedThreads  = 0;

        private object _abortLock;
        private object _executionLock;
        private object _countLock;

        private bool _isExecuting = false;
        private ConcurrentDictionary<ThreadWorkPackage, AutoResetEvent> _abortSignals;
                
        #endregion

        #region Events

        /// <summary>
        /// Called when all Threads have finished executing
        /// </summary>
        public delegate void FinishedProcessing();
        public event FinishedProcessing FinishProcessingEvent;

        /// <summary>
        /// Called when an individual thread has finished executing
        /// </summary>
        protected delegate void ThreadFinished();
        protected event ThreadFinished ThreadExecutionFinished;

        /// <summary>
        /// Called when a thread has been aborted
        /// </summary>
        protected delegate void ThreadAborted();
        protected event ThreadAborted ThreadExecutionAborted;

        #endregion

        #region Constructor

        /// <summary>
        /// The default constructor
        /// Multiple instances of this class can exist and are thread safe
        /// </summary>
        public ThreadPoolManager()
        {            
            ThreadExecutionFinished += new ThreadFinished(ThreadPoolManagerThreadExecutionFinished);
            ThreadExecutionAborted += new ThreadAborted(ThreadPoolManagerThreadExecutionAborted);
           
            _abortSignals  = new ConcurrentDictionary<ThreadWorkPackage, AutoResetEvent>();
            _countLock     = new object();
            _executionLock = new object();
            _abortLock     = new object();
        }              

        #endregion        

        #region Public Properties

        /// <summary>
        /// Number of Aborted Threads
        /// </summary>
        public int AbortedThreads
        {
            get { return _abortedThreads; }
        }

        /// <summary>
        /// Exposed the Executed Count
        /// </summary>
        public int ExecutedThreads
        { 
            get { return _executedThreads; } 
        }
        
        /// <summary>
        /// Exposed the Thread Count
        /// </summary>
        public int TotalThreads 
        { 
            get { return _totalThreads; } 
        }        

        /// <summary>
        /// Retrieves the Execution State
        /// </summary>
        public bool IsExecuting
        {
            get { return _isExecuting; }
        }

        /// <summary>
        /// Used to trigger aborts on child threads
        /// </summary>
        protected ConcurrentDictionary<ThreadWorkPackage, AutoResetEvent> AbortSignals
        {
            get { return _abortSignals; }
        }

        #endregion

        #region Public Functions

        /// <summary>
        /// Used for Wait Signal-free Calls
        /// </summary>
        /// <param name="workItems">[Required] A collection of items to execute</param>
        /// <param name="synchronous">[Required] Whether or not to wait until they have completed</param>
        public void BeginExecution(List<ThreadWorkPackage> workItems, bool synchronous)
        {
            BeginExecution(workItems, synchronous, false, null, String.Empty, -1);
        }

        /// <summary>
        /// Used for Wait Signal-free Calls, allows side-by-side execution
        /// </summary>
        /// <param name="workItems">[Required] A collection of items to execute</param>
        /// <param name="synchronous">[Required] Whether or not to wait until they have completed</param>
        /// <param name="ignoreExecution">[Required] Whether to execute or not, if another set of items are currently being executed
        /// Note that ignoring executing threads will delay triggering of the Finished Processing event, as it increments the total count of threads</param>
        public void BeginExecution(List<ThreadWorkPackage> workItems, bool synchronous, bool ignoreExecution)
        {
            BeginExecution(workItems, synchronous, ignoreExecution, null, String.Empty, -1);
        }

        /// <summary>
        /// Start Processing as a batch
        /// </summary>
        /// <param name="workItems">[Required] A collection of items to execute</param>
        /// <param name="synchronous">[Required] Whether or not to wait until they have completed</param>
        /// <param name="ignoreExecution">[Required] Whether to execute or not, if another set of items are currently being executed
        /// Note that ignoring executing threads will delay triggering of the Finished Processing event, as it increments the total count of threads
        /// </param>
        /// <param name="masterSeries">[Optional] A master abort handle, to abort all packages at once</param>
        /// <param name="masterWait">[Optional] A common descriptor - required when using a master abort signal</param>
        /// <param name="timeOut">[Optional] When to timeout from thread execution - used with master abort</param>
        /// <exception cref="NotSupportedException">
        /// Only one execution batch is supported at any one time.  
        /// Wait for the current execution to finish first.
        /// </exception>
        /// <exception cref="InvalidOperationException">
        /// The arguments passed to BeginExecution must not be null and must contain at least one valid item
        /// </exception>
        /// <exception cref="NotSupportedException">
        /// We use the masterSeries string to determine which threads to trigger an abort for, which is why it is a required field
        /// when a master wait handle is used for aborting package threads
        /// </exception>
        public void BeginExecution(List<ThreadWorkPackage> workItems, bool synchronous, bool ignoreExecution, AutoResetEvent masterWait, string masterSeries, int timeOut)
        {
            BeginExecution(workItems, synchronous, ignoreExecution, masterWait, masterSeries, timeOut, 0, 0);
        }

        /// <summary>
        /// Start Processing as a batch
        /// </summary>
        /// <param name="workItems">[Required] A collection of items to execute</param>
        /// <param name="synchronous">[Required] Whether or not to wait until they have completed</param>
        /// <param name="ignoreExecution">[Required] Whether to execute or not, if another set of items are currently being executed
        /// Note that ignoring executing threads will delay triggering of the Finished Processing event, as it increments the total count of threads
        /// </param>
        /// <param name="masterSeries">[Optional] A master abort handle, to abort all packages at once</param>
        /// <param name="masterWait">[Optional] A common descriptor - required when using a master abort signal</param>
        /// <param name="timeOut">[Optional] When to timeout from thread execution - used with master abort</param>
        /// <param name="maxWorkerThreads">[Optional] Max Worker Threads</param>
        /// <param name="maxCompletionPortThreads">[Optional] Max Completion Port Threads</param>
        /// <exception cref="NotSupportedException">
        /// Only one execution batch is supported at any one time.  
        /// Wait for the current execution to finish first.
        /// </exception>
        /// <exception cref="InvalidOperationException">
        /// The arguments passed to BeginExecution must not be null and must contain at least one valid item
        /// </exception>
        /// <exception cref="NotSupportedException">
        /// We use the masterSeries string to determine which threads to trigger an abort for, which is why it is a required field
        /// when a master wait handle is used for aborting package threads
        /// </exception>
        public void BeginExecution(List<ThreadWorkPackage> workItems, bool synchronous, bool ignoreExecution, AutoResetEvent masterWait, string masterSeries, int timeOut, int maxWorkerThreads, int maxCompletionPortThreads)
        {
            if (workItems == null || workItems.Count() <= 0)
            {
                throw new InvalidOperationException("Work Package must not be null and must contain at least one item");
            }

            try
            {
                // locked to prevent multiple calls
                Monitor.Enter(_executionLock);

                if (_isExecuting && !ignoreExecution)
                {
                    // an instance can only process work batches one at a time, otherwise it would mess with the thread synch
                    throw new NotSupportedException("This instance is already processing");
                }

                if (masterWait != null && String.IsNullOrEmpty(masterSeries))
                {
                    // we use the masterSeries string to determine which threads to trigger an abort for, which is why it is a required field
                    // when a master wait handle is used for aborting package threads
                    throw new NotSupportedException("Usage of a master synchronization object (masterWait) requires a value for the (masterSeries) parameter");
                }

                if (_isExecuting)
                {
                    _totalThreads += workItems.Count();
                }
                else
                {
                    _isExecuting     = true;
                    _totalThreads    = workItems.Count();
                    _executedThreads = 0;
                    _abortedThreads  = 0;
                    _abortSignals    = new ConcurrentDictionary<ThreadWorkPackage, AutoResetEvent>();
                }
                
                bool useMasterThreadAbort = masterWait != null;

                if (useMasterThreadAbort)
                {                    
                    MasterAbort masterAbort = new MasterAbort();

                    masterAbort.Manager = this;
                    masterAbort.MasterSeries = masterSeries;

                    ThreadPool.RegisterWaitForSingleObject(masterWait, WaitOrTimerMasterCallback, masterAbort, timeOut, true);
                }

                //enforce thread poool dimensions, if necessary
                if (maxWorkerThreads > 0 && maxCompletionPortThreads > 0)
                {
                    ThreadPool.SetMaxThreads(maxWorkerThreads, maxCompletionPortThreads);
                }

                foreach (ThreadWorkPackage package in workItems)
                {
                    if (useMasterThreadAbort)
                    {
                        AutoResetEvent waitSignal = new AutoResetEvent(false);
                        _abortSignals.TryAdd(package, waitSignal);
                        
                        package.WaitEvent = waitSignal;
                        package.MasterWaitEvent = masterWait;
                        package.MasterSeries = masterSeries;
                    }

                    //give the package a management context
                    package.Manager = this; 
                    ThreadPool.QueueUserWorkItem(ThreadProc, package);                    
                }

                if (synchronous)
                {
                    while (_executedThreads != _totalThreads)
                    {
                        EventHandlers.HandleAsynchEvent();
                        Thread.Sleep(100);
                    }
                }
            }
            finally
            {
                Monitor.Exit(_executionLock);
            }
        }

        /// <summary>
        /// Start Processing a single work item (Asynchronously)
        /// </summary>        
        /// <param name="workItem"></param>
        /// <param name="synchronous"></param>
        /// <exception cref="ArgumentNullException">
        /// The item passed to BeginExecution must not be null
        /// </exception>
        /// <exception cref="NotSupportedException">
        /// Only one execution batch is supported at any one time.  
        /// Wait for the current execution to finish first.
        /// </exception>
        public void BeginExecution(ThreadWorkPackage workItem, bool synchronous)
        {
            BeginExecution(workItem, synchronous, false);
        }

        /// <summary>
        /// Start Processing a single work item (Asynchronously)
        /// </summary>        
        /// <param name="workItem"></param>
        /// <param name="synchronous"></param>
        /// <param name="ignoreExistingExecution"></param>
        /// <exception cref="ArgumentNullException">
        /// The item passed to BeginExecution must not be null
        /// </exception>
        /// <exception cref="NotSupportedException">
        /// Only one execution batch is supported at any one time.  
        /// Wait for the current execution to finish first.
        /// </exception>
        public void BeginExecution(ThreadWorkPackage workItem, bool synchronous, bool ignoreExistingExecution)
        {
            if (workItem == null)
            {
                throw new ArgumentNullException("Work Package must not be null");
            }

            try
            {
                // locked to prevent multiple calls
                Monitor.Enter(_executionLock);

                if (_isExecuting && !ignoreExistingExecution)
                {
                    // an instance can only process work batches one at a time, otherwise it would mess with the thread synch
                    throw new NotSupportedException("This instance is already processing");
                }

                if (ignoreExistingExecution && _isExecuting)
                {
                    ++_totalThreads;
                }
                else
                {
                    _isExecuting = true;

                    _totalThreads = 1;
                    _executedThreads = 0;
                    _abortedThreads = 0;
                }                

                //give the item a management context
                workItem.Manager = this;
                
                //kick off the thread
                ThreadPool.QueueUserWorkItem(ThreadProc, workItem);

                if (synchronous)
                {
                    while (_executedThreads != _totalThreads)
                    {
                        EventHandlers.HandleAsynchEvent();
                        Thread.Sleep(100);
                    }
                }
            }
            finally
            {
                Monitor.Exit(_executionLock);
            }
        }

        #endregion

        #region Private Functions

        /// <summary>
        /// Removes a Wait Signal (after a thread has finished executing)
        /// </summary>
        /// <param name="package">The Package (Key) to remove</param>
        private void RemoveSignalPackage(ThreadWorkPackage package)
        {
            AutoResetEvent output;
            if (_abortSignals != null && _abortSignals.ContainsKey(package))
            {
                _abortSignals.TryRemove(package, out output);
            }
        }

        /// <summary>
        /// Registers or unregisters the Thread Abort Callbacks
        /// </summary>
        /// <param name="package"></param>
        private static void ManageAbortSignal(ThreadWorkPackage package, bool register)
        {
            if (package.WaitEvent == null)
            {
                return;
            }

            try
            {
                if (register)
                {
                    package.ExecutingThread = Thread.CurrentThread;
                    ThreadPool.RegisterWaitForSingleObject(package.WaitEvent, WaitOrTimerCallback, package, package.WaitTimeout, true);
                }
                else
                {
                    package.Manager.RemoveSignalPackage(package);
                    package.WaitEvent.Dispose();
                }
            }
            catch (Exception ex)
            {
                EventHandlers.HandleExceptionEvent(ex, "ManageAbortSignal", true);
            }
        }

        /// <summary>
        /// Used to Abort Some or All of the Running Threads which have registered with an Abort State Handle
        /// </summary>
        /// <param name="masterSeries"></param>
        private void AbortThreads(string masterSeries)
        {
            Log.WriteLine(LogProviderType.DebugLog, LogPriority.Critical, "Kill Series '" + masterSeries + "'");

            if (_abortSignals == null || _abortSignals.Count() == 0)
            {
                Log.WriteTrace(LogProviderType.FlatFile, LogPriority.Critical, "Master Abort Called, no child threads");
                return;
            }

            foreach (KeyValuePair<ThreadWorkPackage, AutoResetEvent> wait in _abortSignals)
            {
                if (wait.Value != null)
                {
                    if (String.IsNullOrEmpty(masterSeries) || wait.Key.MasterSeries == masterSeries)
                    {                        
                        wait.Value.Set();
                    }
                }
            }
        }

        #endregion

        #region Thread Event Processing

        /// <summary>
        /// Occurs when a Thread has finished processing
        /// </summary>
        private void ThreadPoolManagerThreadExecutionFinished()
        {
            try
            {
                Monitor.Enter(_countLock);
                
                if (_executedThreads < 0)
                {
                    _executedThreads = 1;
                }
                else
                {
                    ++_executedThreads;
                }
           
                if (_executedThreads == _totalThreads)
                {
                    _isExecuting = false; //we're done
                    
                    if (FinishProcessingEvent == null)
                        return;

                    FinishProcessingEvent();
                }
            }
            finally
            {
                Monitor.Exit(_countLock);
            }
        }

        /// <summary>
        /// 
        /// </summary>
        private void ThreadPoolManagerThreadExecutionAborted()
        {
            try
            {
                Monitor.Enter(_abortLock);
                
                if (_abortedThreads < 0)
                {
                    _abortedThreads = 1;
                }
                else
                {
                    ++_abortedThreads;
                }
            }
            finally
            {
                Monitor.Exit(_abortLock);
            } 
        }

        /// <summary>
        /// Called by a Thread When Finished
        /// </summary>
        protected void ThreadExecutionComplete()
        {
            if (ThreadExecutionFinished == null)
                return;

            // Fires an Execution Finished Event
            ThreadExecutionFinished();
        }

        /// <summary>
        /// Called by a Thread When Finished
        /// </summary>
        protected void ThreadExecutionInterrupted()
        {
            if (ThreadExecutionAborted == null)
                return;

            // Fires an Execution Finished Event
            ThreadExecutionAborted();
        }        

        #endregion

        #region Thread Procs

        /// <summary>
        /// Starts Thread Processing - Called Internally
        /// </summary>
        /// <param name="stateInfo"></param>
        private static void ThreadProc(object stateInfo)
        {
            ThreadWorkPackage package = stateInfo as ThreadWorkPackage;
            
            if (package == null || package.Manager == null)
                return;

            try
            {
                if (package.UseWaitEvent)
                {
                    ManageAbortSignal(package, true);
                }

                package.ProcessPackage();                
            }
            //this *should* be caught by the "client" code, but if someone forgets, 
            //we need to handle it to prevent it from bringing down the house
            catch (ThreadInterruptedException e)
            {
                EventHandlers.HandleExceptionEvent(e);
                return;
            }
            catch (ThreadAbortException ex)
            {
                Trace.WriteLine("Thread Abort Called " + package.PackageSeries + " " + package.PackageId);
                EventHandlers.HandleExceptionEvent(ex);
                return;
            }
            catch (Exception x)
            {
                EventHandlers.HandleExceptionEvent(x, " from ThreadPoolManager's Thread Proc", true);
            }
            finally
            {
                if (package != null)
                {
                    if(package.UseWaitEvent)
                    {
                        ManageAbortSignal(package, false);
                    }
                    if (package.Manager != null)
                    {
                        package.Manager.ThreadExecutionComplete();
                    }
                }                
            }
        }        

        /// <summary>
        /// Used to Shutdown Child Threads
        /// </summary>
        /// <param name="stateInfo"></param>
        /// <param name="timedOut"></param>
        private static void WaitOrTimerMasterCallback(Object stateInfo, bool timedOut)
        {
            MasterAbort master = stateInfo as MasterAbort;
            if (master == null)
                return;

            ThreadPoolManager instance = master.Manager;
            string masterSeries = master.MasterSeries;

            if (instance == null)
                return;

            instance.AbortThreads(masterSeries);
        }

        /// <summary>
        /// Aborts a thread
        /// </summary>
        /// <param name="stateInfo"></param>
        /// <param name="timedOut"></param>
        private static void WaitOrTimerCallback(Object stateInfo, bool timedOut)
        {
            ThreadWorkPackage package = stateInfo as ThreadWorkPackage;

            if (package == null || package.Manager == null)
                return;

            try
            {
                if (package.ExecutingThread != null)
                {                    
                    // abort the package
                    package.AbortPackage(timedOut);

                    // give it some time to clean up
                    Thread.Sleep(1000);

                    // call the task abort event
                    //context.ExecutingThread.Abort();
                    package.ExecutingThread.Interrupt();

                    // let the manager know the thread was aborted
                    package.Manager.ThreadExecutionInterrupted();
                }
            }
            catch(Exception ex)
            {
                EventHandlers.HandleExceptionEvent(ex, "WaitOrTimerCallback callback threw an exception");
                return;
            }
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Called to Cleanup the Class Resources
        /// </summary>
        public void Dispose()
        {
            AbortThreads(String.Empty);
        }

        #endregion
    }

    #region MasterAbort Class

    /// <summary>
    /// Used to Wrap the Master Abort Sequence and pass to the WaitOrTimerMasterCallback callback
    /// </summary>
    public class MasterAbort
    {
        public ThreadPoolManager Manager { get; set; }
        public String MasterSeries { get; set; }
    }

    #endregion
}
