﻿// ---------------------------------------------------------------------
// This source file is (C) 2010-2013 Sanders Technology http://www.sanderstechnology.com
// Please review the copyright terms in the Conditions.txt file accompanying this source code
// Author:  Rob Sanders
// Email:   rob.sanders@gmail.com
// Created: Thursday, 10 February 2011, Last Modified: 23 October, 2012
// ---------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace ThreadManager
{
    /// <summary>
    /// Used by the Thread Pool to execution functionality
    /// </summary>
    public class ThreadWorkPackage
    {
        #region Constructor

        /// <summary>
        /// Constructor Takes a Delegate to Ensure it is always set properly
        /// </summary>
        /// <param name="PackageProcessingEvent"></param>
        public ThreadWorkPackage()
        {
        }

        #endregion

        #region Public Properties

        /// <summary>
        /// [Used by the ThreadPoolManager] The Host Thread Pool Manager for Execution
        /// </summary>
        public ThreadPoolManager Manager       { get; set; }
        
        /// <summary>
        /// An individual Object Context
        /// </summary>
        public Object            Context       { get; set; }

        /// <summary>
        /// The Execution Context (Used for UI Synch)
        /// Set to a Wait Context if UseWaitEvent is true
        /// </summary>
        public Object            ExecutionContext { get; set; }

        /// <summary>
        /// [Used by the ThreadPoolManager] Contains the actual executing thread
        /// </summary>
        public Thread            ExecutingThread { get; set; }

        /// <summary>
        /// [Optional] Specifies that the package execution context contains 
        ///            an object used to signal the thread should terminate
        /// </summary>
        public bool              UseWaitEvent { get; set; }

        /// <summary>
        /// [Optional] The syncronization event for this package
        /// </summary>
        public AutoResetEvent    WaitEvent { get; set; }

        /// <summary>
        /// [Used by ThreadPoolManager] The master syncronization event for this package
        /// </summary>
        public AutoResetEvent    MasterWaitEvent { get; set; }

        /// <summary>
        /// [Optional] The Thread Timeout, if a Wait is used
        /// </summary>
        public int WaitTimeout 
        { 
            get 
            { 
                return _waitTimeout; 
            } 
            set 
            { 
                _waitTimeout = value; 
            } 
        }

        /// <summary>
        /// Used to identify this work job within a job order
        /// </summary>
        public int PackageId     
        {
            get
            {
                if (String.IsNullOrEmpty(_packageName))
                {
                    return Int32.MinValue;
                }
                return _packageId;
            }
            set
            {
                _packageId = value;
            }
        }        

        /// <summary>
        /// A common name for a job series (a collection of tasks)
        /// </summary>
        public string PackageSeries 
        { 
            get 
            { 
                if(String.IsNullOrEmpty(_packageName))
                {
                    return String.Empty;
                }
                return _packageName;
            }
            set
            {
                _packageName = value;
            }
        }

        /// <summary>
        /// A common name for a job series (a collection of tasks)
        /// </summary>
        public string MasterSeries
        {
            get
            {
                if (String.IsNullOrEmpty(_masterPackageName))
                {
                    return String.Empty;
                }
                return _masterPackageName;
            }
            set
            {
                _masterPackageName = value;
            }
        }

        #endregion

        #region Events

        public delegate void ExecutionEvent(ThreadWorkPackage package);
        public event ExecutionEvent PackageProcessingEvent;

        public delegate void ExecutionAbortEvent(ThreadWorkPackage package, bool timeOut);
        public event ExecutionAbortEvent PackageAbortEvent;

        /// <summary>
        /// Called by the Thread Pool to begin execution
        /// </summary>
        public void ProcessPackage()
        {
            if (PackageProcessingEvent != null)
            {
                PackageProcessingEvent(this);
            }           
        }

        /// <summary>
        /// Called when the Package's Wait Handle is signalled, 
        /// or if a timeout is set, when timeout occurs
        /// </summary>
        /// <param name="timeOut"></param>
        public void AbortPackage(bool timeOut)
        {
            if (PackageAbortEvent != null)
            {
                PackageAbortEvent(this, timeOut);
            }
        }

        #endregion

        #region Private Variables (Backing Variables)

        private int    _packageId;
        private string _packageName;
        private string _masterPackageName;
        private int    _waitTimeout = -1;

        #endregion
    }
}
